/*
 *  SSL client demonstration program
 *
 *  Copyright (C) 2006-2015, ARM Limited, All Rights Reserved
 *  SPDX-License-Identifier: Apache-2.0
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); you may
 *  not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  This file is part of mbed TLS (https://tls.mbed.org)
 */

#if !defined(MBEDTLS_CONFIG_FILE)
#include "mbedtls/config.h"
#else
#include MBEDTLS_CONFIG_FILE
#endif

#if defined(MBEDTLS_PLATFORM_C)
#include "mbedtls/platform.h"
#else
#include <stdio.h>
#define mbedtls_fprintf    fprintf
#define mbedtls_printf     printf
#endif



#include "mbedtls/net.h"
#include "mbedtls/debug.h"
#include "mbedtls/ssl.h"
#include "mbedtls/entropy.h"
#include "mbedtls/ctr_drbg.h"
#include "mbedtls/error.h"
#include "mbedtls/certs.h"
#include "cmsis_os2.h"
#include "rl_net.h"

#include <string.h>

#define SERVER_PORT "4433"
#define SERVER_NAME "172.16.254.131"
#define GET_REQUEST "GET / HTTP/1.0\r\n\r\n"
#define CIPHERSUITE "TLS-PSK-WITH-AES-128-CCM-8"
#define LINK_DOWN 0
#define LINK_UP 1 

#define MAX_RETRY 3

#define DEBUG_LEVEL 1

static void my_debug( void *ctx, int level,
                      const char *file, int line,
                      const char *str )
{
    ((void) level);

    mbedtls_fprintf( (FILE *) ctx, "%s:%04d: %s", file, line, str );
    fflush(  (FILE *) ctx  );
}

// ----------------------------------------------------------------------------------
#include "stm32f4xx_hal.h"              // Keil::Device:STM32Cube HAL:Common

/* This is the timer structure for use with the timing_get_delay and timing_set_delay functions. 
   We are not using the intermediate delay
*/
struct dtls_timer {
	uint32_t tick_start;
	uint32_t fin_ms;
};


#ifdef RTE_CMSIS_RTOS
extern uint32_t os_time;

uint32_t HAL_GetTick(void) {
  return os_time; 
}
#endif /* RTE_CMSIS_RTOS */

#ifdef RTE_CMSIS_RTOS2_RTX5

/* set timer callback */
void timing_set_delay(void *data, uint32_t int_ms, uint32_t fin_ms)
{
	struct dtls_timer *ctx = (struct dtls_timer *)data;

	ctx->fin_ms = fin_ms;

	if (ctx->fin_ms != 0) {
		ctx->tick_start = osKernelGetTickCount(); 
	}
}

/* get timer callback */
int timing_get_delay(void *data)
{
	struct dtls_timer *ctx = (struct dtls_timer *)data;
	unsigned long elapsed_ms;
	uint32_t tick_now,ticks_per_ms; 
	
	// If timer has been reset then we drop out.
	if (ctx->fin_ms == 0) {
		return -1;
	}
		
  tick_now = osKernelGetTickCount();
	ticks_per_ms=osKernelGetTickFreq() / 1000;

  elapsed_ms = (tick_now - ctx->tick_start) / ticks_per_ms;
	
	if (elapsed_ms >= ctx->fin_ms) {
		return 2;
	}
	
	return 0;
}
#endif /* RTE_CMSIS_RTOS_RTX5 */

uint32_t link_status; 

/* Example implementation of the netETH_Notify() callback to 
 * learn about link status changes. 
*/

void netETH_Notify (uint32_t if_num, netETH_Event event, uint32_t val) {
  NET_ETH_LINK_INFO *info;
     
  switch (event) {
    case netETH_LinkDown:
      mbedtls_printf ("Link is down\n");
		  link_status=LINK_DOWN; 
      break;
    case netETH_LinkUp:
      mbedtls_printf ("Link is up\n");
		  link_status=LINK_UP; 
      info = (NET_ETH_LINK_INFO *)&val;
      switch (info->speed) {
        case 0:
          mbedtls_printf ("10 MBit\n");
          break;
        case 1:
          mbedtls_printf ("100 MBit\n");
          break;
        case 2:
          mbedtls_printf ("1 GBit\n");
          break;
      }
      switch (info->duplex) {
        case 0:
          mbedtls_printf ("Half duplex\n");
          break;
        case 1:
          mbedtls_printf ("Full duplex\n");
          break;
      }
      break;
    case netETH_Wakeup:
      mbedtls_printf ("Wakeup frame received\n");
      break;
    case netETH_TimerAlarm:
      mbedtls_printf ("Timer alarm\n");
      break;
  }
} 

RNG_HandleTypeDef hrng;

/* The entropy_poll function is a callback provided to 
   mbedtls_entropy_add_source().
	 
	 The arguments are: 
	 - 'data' contains a pointer to the RNG_HandleTypeDef 
	   structure
   - 'output' is a buffer of length 'len' allocated by 
	   the calling function
   - 'olen' is a return value indicating how many 
	   random bytes have been placed in the 'output' 
		 buffer. 
	  
*/
int entropy_poll( void *data, unsigned char *output,
                             size_t len, size_t *olen )
{
	RNG_HandleTypeDef *pRNG=(RNG_HandleTypeDef *) data; 
	uint32_t i=0; 
	uint32_t o_rand, n_rand,init=1; 
	HAL_StatusTypeDef ret; 
	
	if (data == NULL) { 
		mbedtls_printf("\n    No pointer to RNG_HandleTypeDef provided in entropy_poll function call ! " );				
		return (MBEDTLS_ERR_ENTROPY_SOURCE_FAILED );
	}		

  memset(output,0x0,len);
	
	/* Only fetch a whole number of words. The STM32F407 RNG produces 
	   random numbers in 32 bit length. 
	*/
	len = (len /4 ) * 4; 

	while (i < len) { 

		/* In the case of a seed error, the generation of random numbers is 
		   interrupted for as long as the SECS bit is '1'. 
		*/
		
		if (__HAL_RNG_GET_FLAG(pRNG, RNG_FLAG_SECS)==SET) {

						mbedtls_printf("\n    HAL_ERROR - RNG_FLAG_SECS (seed error)! " );
						return (HAL_ERROR );
		}

		/* In the case of a clock error, the RNG is no more able to generate random 
		   numbers because the RNG_CLK clock is not correct.
		*/
		
		if (__HAL_RNG_GET_FLAG(pRNG, RNG_FLAG_CECS)==SET) {

						mbedtls_printf("\n    HAL_ERROR - RNG_FLAG_CECS (clock error)! " );
						return (HAL_ERROR );
		}
		
		/* Section 24.4.3 of the reference manual for the 
		   STM32F407 MCU (RM0090) says that the software
		   must check the DRDY bit, i.e. bit position 0 in the 
       RNG_SR register, before reading data from the RNG data 
		   register (RNG_DR). 
			 
			 This check is done in the HAL_RNG_GenerateRandomNumber
			 function.
		*/ 
		ret=HAL_RNG_GenerateRandomNumber(pRNG, &n_rand); 
		
		switch (ret) { 
			case HAL_OK: 
				/* As required by the FIPS PUB 140-2, the first random number 
			     generated after setting the RNGEN bit should not be used, 
				 	 but saved for comparison with the next generated random number. 
				   Each subsequent generated random number has to be compared with 
			     the previously generated number. The test fails if any two 
			     compared numbers are equal (continuous random number generator 
			     test). 
			  */

				if (init==1) { 
					// first random number will be discarded
					init=0; 
				} else {
					// Check whether two consequtive random numbers are equal
					if (o_rand==n_rand) { 
						// error if the two match
						mbedtls_printf("\n    Two random numbers produced the same value ! " );
						return (MBEDTLS_ERR_ENTROPY_SOURCE_FAILED );
					}	else { 
						// everything OK 		
						memcpy(&output[i], &n_rand,4); 
					}
					i+=4; 
				}
				o_rand=n_rand; 
				break; 
			case HAL_ERROR: 
			  mbedtls_printf("\n    HAL_RNG_GenerateRandomNumber() returned HAL_ERROR ! " );
				return (MBEDTLS_ERR_ENTROPY_SOURCE_FAILED );
			default:
			  mbedtls_printf("\n    HAL_RNG_GenerateRandomNumber() returned unknown error ! " );				
				return (MBEDTLS_ERR_ENTROPY_SOURCE_FAILED );
		}
	}

	*olen = i;

	return( 0 );
}

/*
 * recv/send functions that make sure each try returns
 * WANT_READ/WANT_WRITE at least once before sucesseding
 */
static int my_recv( void *ctx, unsigned char *buf, size_t len )
{
    static int first_try = 1;
    int ret;

    if( first_try )
    {
        first_try = 0;
        return( MBEDTLS_ERR_SSL_WANT_READ );
    }

    ret = mbedtls_net_recv( ctx, buf, len );
    if( ret != MBEDTLS_ERR_SSL_WANT_READ )
        first_try = 1; /* Next call will be a new operation */
    return( ret );
}

static int my_send( void *ctx, const unsigned char *buf, size_t len )
{
    static int first_try = 1;
    int ret;

    if( first_try )
    {
        first_try = 0;
        return( MBEDTLS_ERR_SSL_WANT_WRITE );
    }

    ret = mbedtls_net_send( ctx, buf, len );
    if( ret != MBEDTLS_ERR_SSL_WANT_WRITE )
        first_try = 1; /* Next call will be a new operation */
    return( ret );
}

/// System Clock Configuration
void SystemClock_Config(void) {
  RCC_OscInitTypeDef RCC_OscInitStruct;
  RCC_ClkInitTypeDef RCC_ClkInitStruct;

  /* Enable Power Control clock */
  __HAL_RCC_PWR_CLK_ENABLE();

  /* The voltage scaling allows optimizing the power consumption when the
     device is clocked below the maximum system frequency (see datasheet). */
  __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE1);

  /* Enable HSE Oscillator and activate PLL with HSE as source */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSE;
  RCC_OscInitStruct.HSEState = RCC_HSE_ON;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
  RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSE;
  RCC_OscInitStruct.PLL.PLLM = 25;
  RCC_OscInitStruct.PLL.PLLN = 336;
  RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV2;
  RCC_OscInitStruct.PLL.PLLQ = 7;
  HAL_RCC_OscConfig(&RCC_OscInitStruct);

  /* Select PLL as system clock source and configure the HCLK, PCLK1 and PCLK2
     clocks dividers */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_SYSCLK | RCC_CLOCKTYPE_PCLK1 |
                                RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV4;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV2;
  HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_5);
	
}
// ----------------------------------------------------------------------------------


/* RNG init function */
void MX_RNG_Init(void)
{

  hrng.Instance = RNG;
  if (HAL_RNG_Init(&hrng) != HAL_OK)
  {
    mbedtls_printf( " HAL_RNG_Init failed\n");
  }

}

void HAL_RNG_MspInit(RNG_HandleTypeDef* hrng)
{

  if(hrng->Instance==RNG)
  {
  /* USER CODE BEGIN RNG_MspInit 0 */

  /* USER CODE END RNG_MspInit 0 */
    /* Peripheral clock enable */
    __HAL_RCC_RNG_CLK_ENABLE();
  /* USER CODE BEGIN RNG_MspInit 1 */

  /* USER CODE END RNG_MspInit 1 */
  }

}

// Application main thread
static void app_main (void *argument)
{
    int ret, len;
    mbedtls_net_context server_fd;
    unsigned char buf[1024];
    const char *pers = "ssl_client1-2d89f879-0887-4916-8664-30a8c76eaeca"; // device-unique personalization string
	  int force_ciphersuite[2];
	  unsigned char psk[]={0x57, 0x90, 0x84, 0x80, 0x63, 0x19, 0x39, 0x68, 0x30, 0x70, 0x81, 0x68, 0x24, 0x01, 0x36, 0x84, 0x84, 0x26, 0x23, 0x99,0x00};
 	  unsigned char psk_identity[]={0x63,0x6C,0x69,0x65,0x6E,0x74,0x00}; // psk_identity = "client"
	  struct dtls_timer timer; 

		int written, frags, retry_left; 
    
    mbedtls_entropy_context entropy;
    mbedtls_ctr_drbg_context ctr_drbg;
    mbedtls_ssl_context ssl;
    mbedtls_ssl_config conf;

#if defined(MBEDTLS_DEBUG_C)
    mbedtls_debug_set_threshold( DEBUG_LEVEL );
#endif

		link_status=LINK_DOWN; 
    /*
     * Initialize TLS session data
     */
    mbedtls_net_init( &server_fd );
    mbedtls_ssl_init( &ssl );
    mbedtls_ssl_config_init( &conf );

    /*
     * Initialize the RNG
     */
    mbedtls_ctr_drbg_init( &ctr_drbg );
    MX_RNG_Init();

    mbedtls_printf( "\n  . Seeding the random number generator..." );
    fflush( stdout );

    mbedtls_entropy_init( &entropy );
				
		if (( ret = mbedtls_entropy_add_source(&entropy,entropy_poll,(void*)&hrng,MBEDTLS_CTR_DRBG_ENTROPY_LEN,MBEDTLS_ENTROPY_SOURCE_STRONG))!=0)
		{
		    mbedtls_printf( " failed\n  ! mbedtls_entropy_add_source returned %d\n", ret );
        goto exit;
    }

    mbedtls_ctr_drbg_set_prediction_resistance(&ctr_drbg,MBEDTLS_CTR_DRBG_PR_ON);
		
		
    if( ( ret = mbedtls_ctr_drbg_seed( &ctr_drbg, mbedtls_entropy_func, &entropy,
                               (const unsigned char *) pers,
                               strlen( pers ) ) ) != 0 )
    {
        mbedtls_printf( " failed\n  ! mbedtls_ctr_drbg_seed returned %d\n", ret );
        goto exit;
    }

    mbedtls_printf( " ok\n" );


    /*
     * Start the connection
     */
    mbedtls_printf( "  . Connecting to udp/%s/%s...", SERVER_NAME, SERVER_PORT );
    fflush( stdout );
		
    if( ( ret = mbedtls_net_connect( &server_fd, SERVER_NAME,
                                         SERVER_PORT, MBEDTLS_NET_PROTO_UDP ) ) != 0 )
    {
        mbedtls_printf( " failed\n  ! mbedtls_net_connect returned %d\n\n", ret );
        goto exit;
    }

		// osDelay(3000);  // Give networking stack time to initialize
		// 
		while (link_status == LINK_DOWN); 
		
    /*
     * Setup TLS
     */
    mbedtls_printf( "  . Setting up the SSL/TLS structure..." );
    fflush( stdout );

    if( ( ret = mbedtls_ssl_config_defaults( &conf,
                    MBEDTLS_SSL_IS_CLIENT,
                    MBEDTLS_SSL_TRANSPORT_DATAGRAM,
                    MBEDTLS_SSL_PRESET_DEFAULT ) ) != 0 )
    {
        mbedtls_printf( " failed\n  ! mbedtls_ssl_config_defaults returned %d\n\n", ret );
        goto exit;
    }

    mbedtls_printf( " ok\n" );

		// Configure RNG for use with TLS
    mbedtls_ssl_conf_rng( &conf, mbedtls_ctr_drbg_random, &ctr_drbg );
    // Configure debugging support for TLS
    mbedtls_ssl_conf_dbg( &conf, my_debug, stdout );
		// configure timers for retransmissions of lost handshake messags (1 sec initial and 60 secs for max)
		mbedtls_ssl_conf_handshake_timeout(&conf, 1000, 60000);

    // enable anti-replay protection
#if defined(MBEDTLS_SSL_DTLS_ANTI_REPLAY)
		mbedtls_ssl_conf_dtls_anti_replay(&conf, MBEDTLS_SSL_ANTI_REPLAY_ENABLED);
#endif
    // badmac limit for 10 packets
#if defined(MBEDTLS_SSL_DTLS_BADMAC_LIMIT)		
		mbedtls_ssl_conf_dtls_badmac_limit(&conf, 10); 
#endif
		
		// mtu 
		mbedtls_ssl_set_mtu(&ssl,1500);
				
    if( ( ret = mbedtls_ssl_setup( &ssl, &conf ) ) != 0 )
    {
        mbedtls_printf( " failed\n  ! mbedtls_ssl_setup returned %d\n\n", ret );
        goto exit;
    }

		// Configure use of specific ciphersuite
		force_ciphersuite[0] = mbedtls_ssl_get_ciphersuite_id( CIPHERSUITE );

    if( force_ciphersuite[0] != 0 )
    {
				mbedtls_ssl_conf_ciphersuites( &conf, force_ciphersuite );
    }
					
		// Configure send/receive functions
    // mbedtls_ssl_set_bio( &ssl, &server_fd, mbedtls_net_send, mbedtls_net_recv, NULL );
		mbedtls_ssl_set_bio( &ssl, &server_fd, my_send, my_recv, NULL );

		
		// Setting the timer callbacks for DTLS 
    mbedtls_ssl_set_timer_cb( &ssl, &timer, timing_set_delay,
                              timing_get_delay );

	  if( ( ret = mbedtls_ssl_conf_psk( &conf, (const unsigned char *) psk, strlen((const char *)&psk[0]),
                             (const unsigned char *) psk_identity,
                             strlen((const char *)&psk_identity[0])) ) != 0 )
		{
		    mbedtls_printf( " failed\n  ! mbedtls_ssl_conf_psk returned %d\n\n", ret );
		    goto exit;
		}
			
    /*
     * Run TLS handshake protocol
     */
    mbedtls_printf( "  . Performing the TLS handshake..." );
    fflush( stdout );

    while( ( ret = mbedtls_ssl_handshake( &ssl ) ) != 0 )
    {
        if( ret != MBEDTLS_ERR_SSL_WANT_READ && ret != MBEDTLS_ERR_SSL_WANT_WRITE )
        {
            mbedtls_printf( " failed\n  ! mbedtls_ssl_handshake returned -0x%x\n\n", -ret );
            goto exit;
        }
    }

		
    /*
     * Print connection data
     */
		
		mbedtls_printf( " ok\n    [ Protocol is %s ]\n    [ Ciphersuite is %s ]\n",
            mbedtls_ssl_get_version( &ssl ), mbedtls_ssl_get_ciphersuite( &ssl ) );

    if( ( ret = mbedtls_ssl_get_record_expansion( &ssl ) ) >= 0 )
        mbedtls_printf( "    [ Record expansion is %d ]\n", ret );
    else
        mbedtls_printf( "    [ Record expansion is unknown (compression) ]\n" );

    if( ( ret = mbedtls_ssl_get_max_out_record_payload( &ssl ) ) >= 0 )
        mbedtls_printf( "    [ Maximum outgoing record payload in bytes is %d ]\n", ret );
    else
        mbedtls_printf( "    [ Maximum outgoing record payload unknown. ]\n" );

		
#if defined(MBEDTLS_SSL_MAX_FRAGMENT_LENGTH)
    mbedtls_printf( "    [ Maximum fragment length is %u ]\n",
                    (unsigned int) mbedtls_ssl_get_max_frag_len( &ssl ) );
#endif
		
		
		
    mbedtls_printf( " ok\n" );

    /*
     * Write the GET request
     */
    retry_left = MAX_RETRY;

send_request:
		
    mbedtls_printf( "  > Write to server:" );
    fflush( stdout );

    len = sprintf( (char *) buf, GET_REQUEST );

    do ret = mbedtls_ssl_write( &ssl, buf, len );
    while( ret == MBEDTLS_ERR_SSL_WANT_READ ||
           ret == MBEDTLS_ERR_SSL_WANT_WRITE );

    if( ret < 0 )
    {
       mbedtls_printf( " failed\n  ! mbedtls_ssl_write returned %d\n\n", ret );
       goto exit;
    }

    frags = 1;
    written = ret;

    buf[written] = '\0';
    mbedtls_printf( " %d bytes written in %d fragments\n\n%s\n", written, frags, (char *) buf );
		
    /*
     * Read the HTTP response
     */
    mbedtls_printf( "  < Read from server:" );
    fflush( stdout );

    len = sizeof( buf ) - 1;
    memset( buf, 0, sizeof( buf ) );

    do ret = mbedtls_ssl_read( &ssl, buf, len );
    while( ret == MBEDTLS_ERR_SSL_WANT_READ ||
           ret == MBEDTLS_ERR_SSL_WANT_WRITE );

   if( ret <= 0 )
   {
       switch( ret )
       {
           case MBEDTLS_ERR_SSL_TIMEOUT:
                mbedtls_printf( " timeout\n" );
                if( retry_left-- > 0 )
                    goto send_request;
                goto exit;

           case MBEDTLS_ERR_SSL_PEER_CLOSE_NOTIFY:
                mbedtls_printf( " connection was closed gracefully\n" );
                ret = 0;
                goto close_notify;

           default:
                mbedtls_printf( " mbedtls_ssl_read returned -0x%x\n", -ret );
                goto exit;
      }
   }

   len = ret;
   buf[len] = '\0';
   mbedtls_printf( " %d bytes read\n\n%s", len, (char *) buf );
   ret = 0;

	 osDelay(3000);  // Delay before closing connection

    /*
     * Done, cleanly close the connection
     */
close_notify:
    mbedtls_printf( "  . Closing the connection..." );
    fflush( stdout );

	  // Send close notify message.  
    do ret = mbedtls_ssl_close_notify( &ssl );
    while( ret == MBEDTLS_ERR_SSL_WANT_WRITE );
    ret = 0;

    mbedtls_printf( " done\n" );

exit:

#ifdef MBEDTLS_ERROR_C
    if( ret != 0 )
    {
        char error_buf[100];
        mbedtls_strerror( ret, error_buf, 100 );
        mbedtls_printf("Last error was: %d - %s\n\n", ret, error_buf );
    }
#endif

    mbedtls_net_free( &server_fd );

    mbedtls_ssl_free( &ssl );
    mbedtls_ssl_config_free( &conf );
    mbedtls_ctr_drbg_free( &ctr_drbg );
    mbedtls_entropy_free( &entropy );

    return;
}

int main (void) {
  static uint64_t stack[8192/8];
  static const osThreadAttr_t attr = {
    .stack_mem  = stack,
    .stack_size = sizeof(stack)
  };

  // System Initialization
  HAL_Init();
  SystemClock_Config ();
  SystemCoreClockUpdate();

  osKernelInitialize();
  osThreadNew(app_main, NULL, &attr);
  osKernelStart();

  for (;;) {}
}

